#include "heap.h"

#include <limits>

using namespace std;

#ifndef MAX_HEAP
#define MIN_HEAP
#endif

CHeap::CHeap(void)
: m_heapSize(0)
{
	m_root = new CHeapNode();
	m_root->setParent(m_root);

	m_lastNode = m_root;
}

CHeap::~CHeap(void)
{
    CHeapNode* lastNode = m_lastNode;
	CHeapNode* node;
	while(lastNode != NULL)
	{
		node = lastNode->getPrevious();
		delete lastNode;
		lastNode = node;
	}
}

CHeapNode* CHeap::getRoot(void) const
{
	return m_root;
}

long CHeap::getHeapSize(void) const
{
	return m_heapSize;
}

CHeapNode* CHeap::top(void) const
{
	return m_root->getLeftChild();
}

void CHeap::pop(void)
{
	deleteNode(m_root->getLeftChild());
}

void CHeap::clear(void)
{
    CHeapNode* lastNode = m_lastNode;
	CHeapNode* node;
	while(lastNode != m_root)
	{
		node = lastNode->getPrevious();
		delete lastNode;
		lastNode = node;
	}

	m_root->setLeftChild(NULL);
	m_root->setPrevious(NULL);
	m_root->setNext(NULL);

	m_lastNode = m_root;
	m_heapSize = 0;
}

void CHeap::insertNode(double key, void* data)
{
	CHeapNode* pNode = new CHeapNode(key, data);
  m_heapSize ++;

	CHeapNode* parent = getInsertPoint();
	if(parent->getLeftChild() == NULL)
	{
		parent->setLeftChild(pNode);
	}
	else
	{
		parent->setRightChild(pNode);
	}
	pNode->setParent(parent);

	CHeapNode* node = pNode;
	while(parent != m_root && isWrongPos(parent, node))
	{
		parent->swapData(node);
		node = parent;
		parent = parent->getParent();
	}

	m_lastNode->setNext(pNode);
	pNode->setPrevious(m_lastNode);
	m_lastNode = pNode;
}

void CHeap::deleteNode(CHeapNode* pNode)
{
  bool heapifyUpwards = isWrongPos(pNode, m_lastNode);
	bool heapifyNeeded = pNode != m_lastNode;

  pNode->swapData(m_lastNode);

  CHeapNode* node = m_lastNode->getParent();
	if(m_lastNode->isLeftChild())
	{
		node->setLeftChild(NULL);
	}
	else
	{
		node->setRightChild(NULL);
	}

	node = m_lastNode;
	m_lastNode = m_lastNode->getPrevious();
	m_lastNode->setNext(NULL);

    delete node;

    m_heapSize --;

	if(heapifyNeeded)
	{
		heapify(pNode, heapifyUpwards);
	}
}

void CHeap::dump(ostream& os, CHeapNode* pNode, bool details) const
{
	if(pNode == m_root)
	{
		pNode = m_root->getLeftChild();
	}
	if(pNode != NULL)
	{
		pNode->dump(os, 0, details);
	}
}

bool CHeap::isWrongPos(CHeapNode* parent, CHeapNode* child)
{
#ifdef MIN_HEAP
	// do swap if parent.key > child.key
	return(parent->compareKey(child->getKey()) == 1);
#else
	// do swap if parent.key < child.key
	return(parent->compareKey(child->getKey()) == -1);
#endif
}

void CHeap::heapify(CHeapNode* pNode, bool upwards)
{
    if(upwards)
    {
        CHeapNode* cur = pNode;
        CHeapNode* parent = cur->getParent();
        bool cont = true;

        while(parent != m_root && cont)
        {
            if(isWrongPos(parent, cur))
            {
                cur->swapData(parent);

                cur = parent;
                parent = parent->getParent();
            }
            else
            {
                cont = false;
            }
        }
    }
    else
    {
    	CHeapNode* node = pNode;
	    CHeapNode* child = pNode->getLeftChild();

    	if(child != NULL)
	    {
		    if(isWrongPos(node, child))
    		{
	    		node = child;
		    }
    		child = pNode->getRightChild();
	    	if(child != NULL && isWrongPos(node, child))
		    {
			    node = child;
    		}
	    }

    	if(node != pNode)
	    {
		    pNode->swapData(node);

            heapify(node, false);
	    }
     }
}

CHeapNode* CHeap::getInsertPoint(void)
{
	if(m_lastNode->getParent() == m_root)
	{
		return m_lastNode;
	}
	else
	{
		if(m_lastNode->isLeftChild())
		{
			return m_lastNode->getParent();
		}
		else
		{
			return m_lastNode->getParent()->getNext();
		}
	}
}
