#ifndef NODE_H
#define NODE_H

#include <iostream>
#include "mystack.h"
#include <vector>
#include <limits>

#include "rtree.h"
#include "rectangle.h"

using namespace std;


class CRTree;
class CIndexNode;
class CLeafNode;

class CEntry  // all data points are sotred in CEntry
{
public:
  long m_id;      // id of this data point
  CRectangle* m_mbr;  // mbr of this data point

  long m_dataCnt;   // aggregate number of data points
  double m_p; //  aggregate probabilities of data points

  CEntry(long id, CRectangle* mbr, long dataCnt, double prob)
      : m_id(id),
      m_mbr(mbr),
      m_dataCnt(dataCnt),
      m_p(prob)
  {}

  CEntry( CEntry* pEnt )
      : m_id( -pEnt->m_id ),
      m_dataCnt( pEnt->m_dataCnt ),
      m_p( pEnt->m_p)
  {
    m_mbr = new CRectangle( pEnt->getRect() );
  }

  // get the number of aggregation
  long getAggCnt() const { return  m_dataCnt;}
  double getProb() const { return m_p;}
  CRectangle* getRect() const {return m_mbr;}

  virtual ~CEntry(void)
  {
    if ( m_mbr != NULL )  delete m_mbr;
  }

  friend class CRTree;
  friend class CNode;
  friend class CIndexNode;
  friend class CLeafNode;
};

class CNode : public CEntry
{
public:
  // m_id: the id of the block in the disk file;
  // -1 means the node is new and not exists in the disk file
  // long m_id;

  // m_level: level of this node in the r-tree
  // 0: leaf node; > 0: index node
  long m_level;

  // m_mbr: the minimum boxing rectangle of this node
  // CRectangle* m_mbr;

  // m_rtree: the reference of the r-tree
  CRTree* m_rtree;

  // the actual number of child entries in this node
  long m_size;

  // aggregate data information part
  // m_dataCnt: the number of data points rooted by this node
  //long m_dataCnt;

  // m_fm: the fm sketch of all data points rooted by this node
  // CFMData* m_fm;

  class CReinsertEntry
  {
  public:
    long m_index;
    double m_dist;

    CReinsertEntry(long index, double dist)
        : m_index(index),
        m_dist(dist)
    {
    }

    static int compare(const void* v1, const void* v2)
    {
      CReinsertEntry* e1 = *(CReinsertEntry**)v1;
      CReinsertEntry* e2 = *(CReinsertEntry**)v2;

      if ( e1->m_dist < e2->m_dist )
      {
        return -1;
      }
      else if ( e1->m_dist > e2->m_dist)
      {
        return 1;
      }
      else
      {
        return 0;
      }


      /*      if ( e1->m_dist < e2->m_dist - numeric_limits<double>::epsilon() )
            {
              return 1;
            }
            else if(e1->m_dist > e2->m_dist + numeric_limits<double>::epsilon())
            {
              return -1;
            }
            else
            {
              return 0;
            }*/

    }
  };

  class CSplitEntry
  {
  public:
    long m_index;
    CRectangle* m_mbr;

    static long m_sortDim;

    CSplitEntry(long index, CRectangle* mbr)
        : m_index(index),
        m_mbr(mbr)
    {
    }

    static int compareLow(const void* v1, const void* v2)
    {
      CSplitEntry* e1 = *(CSplitEntry**)v1;
      CSplitEntry* e2 = *(CSplitEntry**)v2;

      if (e1->m_mbr->getLow(CSplitEntry::m_sortDim) <
              e2->m_mbr->getLow(CSplitEntry::m_sortDim))
      {
        return -1;
      }
      else if (e1->m_mbr->getLow(CSplitEntry::m_sortDim) >
               e2->m_mbr->getLow(CSplitEntry::m_sortDim))
      {
        return 1;
      }
      else
      {
        return 0;
      }
    }

    static int compareHigh(const void* v1, const void* v2)
    {
      CSplitEntry* e1 = *(CSplitEntry**)v1;
      CSplitEntry* e2 = *(CSplitEntry**)v2;

      if (e1->m_mbr->getHigh(CSplitEntry::m_sortDim) <
              e2->m_mbr->getHigh(CSplitEntry::m_sortDim))
      {
        return -1;
      }
      else if (e1->m_mbr->getHigh(CSplitEntry::m_sortDim) >
               e2->m_mbr->getHigh(CSplitEntry::m_sortDim))
      {
        return 1;
      }
      else
      {
        return 0;
      }
    }
  };

public:
  // create a node with specified 'id' (later read content from the file)
  CNode(long id);
  // create a node in memory
  CNode(long level, CRTree* rtree);

  virtual ~CNode(void);

  bool isLeaf(void);
  long getNumChild(void) const { return m_size; };

  // getEntry: get the child entry specified by 'index'
  virtual CEntry* getEntry(long index) = 0;

  virtual long getBytesLength(void) = 0;
  virtual void toBytes(long& len, char** data) = 0;
  virtual void fromBytes(char* data) = 0;
  virtual void load(void) = 0;

private:
  virtual CNode* chooseSubtree(CRectangle* mbr, long level,
                               CStack<CNode*>& path) = 0;

  // insertData: insert a new data point, the returned value indicates
  // whether the mbr should be updated
  bool insertData(CEntry* e, CStack<CNode*>& path, bool* overflow);

  // insertEntry: normally insert a child node
  void insertEntry(CEntry* e);

  // adjustTree: update nodes along the path
  // 'recompute': indicate whether m_dataCnt and m_fm should be recomputed
  void adjustTree(CNode* node, CRectangle* oldMBR, CStack<CNode*>& path,
                  CEntry* e, bool recompute);

  // adjustTree: update nodes along the path
  void adjustTree(CNode* curNode, CNode* newNode, CStack<CNode*>& path,
                  bool* overflow);

  // chooseReinsertNode: partition children before splitting
  void chooseReinsertNode(CRectangle* mbr,
                          vector<long>& reinsert, vector<long>& remain);

  // split: do split
  CNode* split(CEntry* e);

  void dump(ostream& os);


  friend class CRTree;
  friend class CIndexNode;
  friend class CLeafNode;
};

#endif
