#ifndef RECTANGLE_H
#define RECTANGLE_H

#include <iostream>
#include "utility.h"
#include "CEquation.h"
#include <cfloat>
#include <list>
#include <assert.h>



using namespace std;

// #ifndef DIMENSION
// #define DIMENSION 2
// #endif

#define MAX_FLOAT FLT_MAX

class CRectangle
{
public:
  double m_low[DIMENSION];
  double m_high[DIMENSION];

  static double m_minCoord[DIMENSION];
  static double m_maxCoord[DIMENSION];
  
  bool isEmpty;

public:
  // create a empty rectangle: lower-left (higher-right) corner
  // is maximum (minimum)
  CRectangle(void);

  // create a point defined by 'coord'
  CRectangle(double* coord);

  // create a rectangle defined by (low, high)
  CRectangle(double* low, double* high);

  // create a rectangle defined by 'mbr'
  CRectangle(CRectangle* mbr);
  
  // Create a rectangle of side_length such that point "center" lies at its center
  CRectangle(CPoint* center, double side_length);
  
  // Works only for two dimensional data: Get rectangle corresponding to [i][j] cell
  CRectangle(int i, int j);

  virtual ~CRectangle(void);

  // getLow: get one coordinate of the lower-left corner
  double* getLow(void);
  double getLow(long index);

  //getHigh: get one coordinate of the higher-right corner
  double* getHigh(void);
  double getHigh(long index);

  // initialize the minimum and maximum corner point
  static void initMinMaxCoord(void);

  // combine: enlarge m_mbr to cover 'mbr'
  void combine(CRectangle* mbr);
  
  // Enlarge this->rectangle to cover point p
  void combine(CPoint* p);

  // combine: calculate a mbr which cover mbrA and mbrB
  void combine(CRectangle* mbrA, CRectangle* mbrB);

  // getArea: return area of this mbr
  double getArea(void);

  // getIntersectArea: calculate the area of insertion of 'mbr'
  double getIntersectArea(CRectangle* mbr);

  // isContain: check whether contians 'mbr'
  bool isContain(CRectangle* mbr);

  // isTouch: check whether touches 'mbr'
  bool isTouch(CRectangle* mbr);

  // clone: copy 'mbr'
  void clone(CRectangle* mbr);
  void clone(double* low, double* high);

  // setEmpty: set lower-left (higher-right) corner to be maximum (minimum)
  void setEmpty(void);
  
  // directly change the MBR
  void setMBR(double* low );
  void setMBR(double* low, double* high );

  // getCenter: calculate the center of this rectangle
  void getCenter(CRectangle* center);

  // getDist: calculate the distance to mbr, here, both 'this' and 'mbr'
  // are points
  double getDist(CRectangle* mbr);

  // getMargin: calculate perimeter
  double getPerimeter(void);

  void dump(ostream& os);

  double getXY();
  double getMXY(); // x*y

  // get the x+y+.. for low or hight coord of the mbr
  double getLXY() const;
  double getHXY() const;

  // return the L2 distance to the ( 0, 0)
  double getL2XY() const;
  double getLfacXY( long factor ) const;
  
  double getMaxHigh() const;
  double getMinLow() const;

  // get relative position between rect and a point
  /*
  4 | 2 | 1
  -----------
  5 | 3 | 2
  ------------
  6 | 5 | 4

  (same to multi-dimensional case)
  1: whole rect dominates point
  2: partial rect dominates point
  3: unknown
  4: no dominate relationship
  5: point dominates partial rect
  6: point dominates whole rect
  */

  static long relativePos(CRectangle* rect, const double* coord);


  // get relative position between two points
  /*
  |
  2 | 1
  -------
  3 | 2
  |
  (same to multi-dimensional case)
  1: query dominates data
  2: no dominate relationship
  3: data dominates query
  */
  static long relativePos(const double* point, const double* coord);
  
  void resetRectangle(void);
  
  // Trim and return part of this->rectangle that cannot be pruned;
  // If the whole rectangle is prunned, the function returns true
  // The unpruned rectangle is returned in mbr
  bool bisectorTrim(CEquation* equation, CRectangle* mbr);
  
  // Prunes this->rect using Q and mbr rectangles; 
  // returns true if the whole rectangle is trimmed;
  bool pruneByMBR(CRectangle* Q,CRectangle* mbr);
  
  
  // returns true if the whole rectangle is pruned
  // otherwise returns false but doesn't prune any part of this->rectangle
  bool isPrunedByMBR(CRectangle* Q, CRectangle* mbr);
  
  
  // Prunes this->rect using instance q and mbr rectangles; 
  // returns true if the whole rectangle is trimmed;
  bool pruneByInstance(CPoint* q, CRectangle* mbr);
  
  // print rectangle in [low_i,high_i] X [low_j,high_j] format
  void print(void);
  
  // returns all corners of this->rectangle into 'corners' 
  void getCorners(CPoint** corners);
  
  // returns true if mbr is intersected with this->rectangle
  bool isIntersected(CRectangle* mbr);
  
  // returns true if point p intersects this->rectangle
  bool isIntersected(CPoint* p);

  
  // returns the minimum distance between mbr and this->rectangle
  double getMinDist(CRectangle* mbr);
  
  // returns the maximum distance between mbr and this->rectangle
  double getMaxDist(CRectangle* mbr);
  
  
  // returns the square of diagonal of mbr
  double getDiagonal(void);
  
  // returns the maximum distance between this->rectangle and point
  double getMaxDist(CPoint* p);
  
  // returns the minimum distance between p and this->rectangle
  double getMinDist(CPoint* p);
  
  
  // returns true if this->rectangle is equal to mbr
  bool equals(CRectangle* mbr);
  
  // prunes this->rec by dominance pruning using Q and mbr;
  
  bool pruneByDominance(CRectangle* mbr, CPoint* bitMapMBR, CPoint* domPoint);
  
  // returns true rectangle mbr and this->rectangle overlap in any dimension i
  bool isOverlapped(CRectangle* mbr);
  
  // return the closest corner of this->rectangle from the rectangle mbr
  // NOTE this method only works if one rectangle dominates other
  CPoint* getClosestCorner(CRectangle* mbr);


  // return the closest corner of this->rectangle from the rectangle mbr
  // NOTE this method only works if one rectangle dominates other
  CPoint* getClosestCorner(CPoint* p);

   // return the furthest corner of this->rectangle from the point p
  CPoint* getFurthestCorner(CPoint* p);
  
  // returns a point that stores 1 or zero for each dimension i
  // for a dimension i, if mbr in dimension i has greater values than this->rec in i the bitMap[i] is set to
  // 2. If mbr in ith dimension has smaller values than this->rec bitMap[i] is set to 1
  CPoint* getDominanceBitMap(CRectangle* mbr);
  
  // Get the useful bisectors between Q and mbr
  // Before calling this function make sure that Q and mbr don't intersect each other
  int getBisectors(CRectangle* Q, CRectangle* mbr, list<CEquation*> *li);
  
  // older version of getBisectors; delete one of these two after experimenting
  int getBisectors2(CRectangle* Q, CRectangle* mbr, list<CEquation*> *li);
  
  // Get the closest point of this->rectangle from the point p
  CPoint* getClosestPoint(CPoint* p);
  
  // checking
  void passList(list<CEquation*> l);
  
  // Prune this->rec using equations present in bisectorsList
  bool pruneByEquations(CRectangle* Q, list<CEquation*> bisectorsList);
  
  
  // point p is a corner of this->rectangle
  // print the corner in (L,H,L) format where L stands for low and H stands for high
  void printLH(CPoint* p);
  
  // Return true if the corner of Q and mbr are opposite. e.g; if Q corner is (LHL), its opposite corner is (HLH)
  bool isOppositeCorner(CRectangle* Q, CRectangle* mbr, CPoint* q, CPoint* p);
  
  // get the opposite corner of q in mbr (see the definition of opposite corner in above function)
  // more precisely; if q in Q is [LHL], then its opposite corner in mbr is [HLH]
  CPoint* getOppositeCorner(CRectangle* Q, CRectangle* mbr, CPoint* q);
  
  
  // return the frontier point between this->rectangle and the mbr
  CPoint* getFrontierPoint(CRectangle* mbr);
  
  // Blow the size to infinity
  void blowHuge(void);
  
  // Load the center into center point
  void getCenter(CPoint* center);
  
  // Changes the rectangle so that the pointt "center" lies at its center
  // if the rectangle coordinates go out of the boundary, adjust them
  void pointCenteric(CPoint* center, double side_length);
  
  // Expand the rectangle so that it covers whole space (e.g; o - 1)
  void coverWholeSpace();
  
  
  // Get max and min angle of point p
  void getMaxMinAngle( double* minAngle, double* maxAngle, CPoint* p); 
 
  
  // returns true if the rectangle is a cell in the grid
  bool isCell();
  
  
  // return the ith child of this->rec 
  // USED ONLY FOR GRID
  // i corresponds to ith coordinate
  CRectangle* getChild(int index);
  
  // return the cell id of this->rec
  int getCellID();
  
  // return true if point p is contained by rectangle
  bool contains(CPoint* p);

//Check in which dimention the points overlap 
  int getoverlap(CPoint* p);

  // Treats lower left corner of this->rect as a point and computes distance from p
  // It is used when rect is actually a point
  double getPointDist(CPoint* p);
  
  // returns true if the rectangle lies or overlaps with the area within the angle range of q
  bool liesInAngleRange(CPoint* q, double start_angle, double end_angle);
  
    // if this->rectangle is pruned by eqn, then return true
  bool prunedByBisector(CEquation* eqn);
  
  float MinDistAtAngle(CPoint* p, float angle);
  
  
  //Get union of rectangle
  // return false if the newRect is empty
  bool unoinRec(CRectangle* rect , CRectangle* newRect );
  
  //If this is empty return true otherwise return false; 
  bool checkEmpty();
  
};

#endif
