#include <iostream>
#include <cstdlib>
#include <fstream>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <algorithm>
#include <queue>
#include <list>
#include <set>
#include <stdio.h>
#include "indexnode.h"
#include "leafnode.h"
#include "rtree.h"
#include "CEquation.h"
#include "cpoint.h"
#include "rectangle.h"
#include "utility.h"
#include "heap.h"
#include "heapnode.h"
#include "hilbert.h"
#include <stack>
#include <map>
#include <inttypes.h>


#define RELEASE
#define NDEBUG

//#include <assert.h>

//#define TPLEnforce // This flag enables the optimization in TPL (e.g., kTrim is repeatedly called as long as the input entry is not the same as the trimmed entry). 

//#define cmpTPLPruning // This flag is to be used for comparing the filtering power of TPL and TPL++ (as in Fig. 5 of the paper). TPLEnforce must be disabled if this flag is to be used. 



#define InfZone_opt  // This flag must be enabled for all experiments



typedef struct minObject
{
   int objID;
   double key;
   CPoint* value;
  
}minObject;

typedef struct partition
{
   set<double> upperArc;
   set<double> lowerArc;
   double k_LowerArc;
   double k_UpperArc;
   double bouderyArc;
   list<minObject*> minList; 

}Partition;


typedef struct infobj{
   int objID;
   CPoint* loc;
   int numVertices;
   CEquation* eq;
   double distFromQ;
   CPoint* cent; // center of line joining  q and infObj
   bool checkCenter; // if the dist from center to be checked
   double hilbertValue;
  
   infobj()
   {
     checkCenter = true;
     numVertices = 0;
   }
    
}infObj;

typedef struct recentry{
   CRectangle* rec;
   int counter;
    
   recentry()
   {
     rec = new CRectangle();
     counter = 0;   
   }
}recEntry;

typedef struct ext{
   double min;
   double max;
   ext()
   {
     min = 10;
     max = -10;
   }
}extremes;

typedef struct vertex{
   int counter;
   CPoint* loc;
   infObj* obj1;
   infObj* obj2;
   double distFromQ;
   double angleFromQ; // to be used only in vPolygon (in other structures its value is not maintained)
  
  // for debugging purpose
   int objID1;
   int objID2;
  
   vertex()
   {
     objID1 = 0;
     objID2 = 0;
     counter = 0;
     distFromQ = 0;
     angleFromQ=0;
   } 
}Vertex;


typedef struct counterObject
{
   int counter;
   CNode* node;
}counterObject;


typedef struct TPLCand
{
   int counter ;
   int candID;
   CPoint* p;
   list<CNode*> toVisit;
}tplCand;


typedef struct query{
   CPoint* loc;
   CRectangle* rec;   // same as query point (used for BOOST)
   CRectangle* approxRegion; // Approximate SR (for FINCH) or InfZone
   list<infObj*> infList;
   list<Vertex*> vList;
   list<CPoint*> vList_Special; // vertices with counter = k-1 
   vector<Vertex*> vPolygon; // vPolygon should use a vector instead of a list to be able to do binary search in logarithmic time
   list<CPoint*> convexChain; // Compute the convex vertices and copy to vPolygon for logarithmic containment 
   set<int> RkNN;
   double minRadius; // radius of minimum approximation
   double maxRadius; // radius of maximum approximation 
   set<int> BruteRkNN;
   map<int, CPoint*> candidates;
   list<TPLCand*> candList;   // candidates set for TPL 
   list<CNode*> Prfn;      // Pruned point for refinement for TPL
   list<CNode*> Nrfn;      // Pruned node for refinement for TPL
  
   Partition* partitions;
  
   double maxUpper ;   // MaxUpper and MinUpper
   double minUpper ;   
   double minLower;
   query()
   {
     minRadius = 0;
     maxRadius = 0;
     maxUpper = 50000;    
     minUpper = 0;
   }
    
}Query;

typedef struct object{
   CPoint* loc;              // current location
   CPoint* old_loc;          // last seen location
} Object;

class gridRecord {
  public:  
    int objID;
    CRectangle* rec;
    float dist;
    bool isObject;
  
};


// use for sorting point list
bool sortOnDimensionalValue (CPoint* a, CPoint* b);

// load queries (e.g., locations) from RTree into the query table
void loadQueries();
float RandomEqual(float min, float max);
void printQuery(int i);
void printInfObj(int qID);

// initialize the influence zone as the data space
void initializeQuery(int i);


// add the influence object to the query and reflect its effect. e.g.,
// add new intersections, update counters, remove intersections
void addInfObj(int objID, CPoint* point, int qID);

// returns true if the point p lies within the unit square space
bool withinDataSpace(CPoint* p);

// check how many bisectors prune it
int computeCounter(CPoint* p, int qID);

void printVertices(int qID);

// for the bisector of new point, update the counters of all intersection points
// that lie on the other side of bisector. Also remove the intersection points
// that become invalid
void updateCountersAndRemoveIntersections(CPoint* point, int qID);

// prune the rectangle if it lies outside all of the circles
bool prunedByCircles(CRectangle* rec,int qID);

void computeInfluenceZone(int qID);

// computes the number of bisectors
void computeNumOfBisectors(int qID);

// estimate the area by considering it as a circle
void estimateInfZoneArea(int qID);

// compute exact area of influnce zone by converting it into triangles
void computeAreaOfInfZone(int qID);

// get bisectors of the influence zone in set bisecs
void getBisecsOfInfZone(int qID,set<int> *bisecs);

// use the line of pList and get the area by triangulating
double getAreaOfEdges(int qID, int objID, list<CPoint*> *pList);

// if p1p2 is a valid boundary edge., triangulate it and calculate the area
double getAreaOfOneEdge(int qID, int objID, CPoint* p1, CPoint* p2);

void printPList(list<CPoint*> *pList);

// get area of the triangle wit vertices p1,p2 and p3
double getTriangleArea(CPoint* p1, CPoint* p2, CPoint* p3);

double computeAreaBySampling(int qID);

// short list the points that will be used for convex hull computation
// e.g., remove the intersections with counter <(k-1)  and keep extreme
// points for boundary vertices
void shortlistPoints(int qID, list<CPoint*> *pList);

// if lies at data space boundary [0,0] X [1,1]
bool liesAtDataSpaceBoundary(CPoint* p);

// for each boundary, keep the extreme points
void keepExtremeBoundaryPoints(list<CPoint*> *boundaryPoints);
// update min max value at index
void updateExtremes(extremes* *pArray, double val, int index);

void printPoints(list<CPoint*> *boundaryPoints);

void deleteVertex(Vertex* v);

void deleteminObj( minObject* x);

void deleteTPLCands( TPLCand* x);
// if the intersection point i lies between p1, and p2
bool intersectionExists(CPoint* i, CPoint* p1, CPoint* p2);

// release the memory of points
void deletePoints(CPoint** points);

// check if the interval (a1,a2) and (b1,b2) intersect with each other
bool edgesIntersect(CPoint* a1, CPoint* a2, CPoint* b1, CPoint* b2);




// if less than k objects within range from rec, return true
bool lessThanKInRange(CPoint* p, double range , int qID );


void printRkNN(int qID);
//void  writetoFile( int qID);

void  INFZONE_BRNN_refinement(int qID);

void INFZONE_BRNN(int qID);

// if rectangle is outside influence zone, return true;
bool outsideInfZone(CRectangle* rec, int qID);

// if p is outside influence zone, return 1 or 2, otherwise reutn 0;
int outsideInfZone(CPoint* p, int qID);
// monochromatic
void INFZONE_MONO(int qID);

void  INFZONE_MONO_refinement(int qID);




bool overlapsInfZone(CRectangle* R, int qID);
bool edgeIntersectsInfZone(CPoint* a1, CPoint* a2,int qID);

//return true if found in buffer, otherwise return false and insert it in buffer after evicting some entry
bool foundInBuffer(int ID);

// avg influence objects
double getAvgInfObj();

double getAvgRkNN();
void printStats(ostream& os);
void printTheoreticalResults(ostream& os);



bool edgeIntersectsPolygon2(CPoint* a1, CPoint* a2,int qID);
bool edgesIntersect2(CPoint* a1, CPoint* a2, CPoint* b1, CPoint* b2);
bool intersectionExists2(CPoint* c, CPoint* p1, CPoint* p2);
bool pointIsInSR2(CPoint* p, int qID);


// get avg of actual area of all queries
double getAvgExactArea();
double getTheoreticalRkNN();

// TYPE = 0 means get pruning phase IO, TYPE=1 means get refinement phase (user tree)
double getTheoreticalIOs(int TYPE);

double getExpectedSize();


void resetQuery(int qID);
void deleteInfObj(infObj* x);

double getAvgVertices();
int getMaxVertices();



void constructPolygon(int qID);
bool sortVerticesOnAngle (Vertex* a, Vertex* b);
void printPolygon(int qID);
bool correctPolygon(int qID);
bool correctCounters(int qID);
bool liesOutOfPolygon(CPoint* p, int qID);
infobj* getCommonObj(Vertex* v1, Vertex* v2);

bool liesOutOfSR(CPoint* p, int qID);





//================SLICE=======================//

void SLICE_BRNN( int qID ,  int pNum);
void initQuerySLICE( int qID );
void SLICE_filtering( int qID );
void updateMinList(int qID);
bool FacilityPruned(CRectangle* rect, int qID  , int node_ID);



void updateZone( CPoint* cp , int qID , int mid);
void updateArc( double lowerArc , double upperArc  , int pNo , int qID , CPoint* cp , int mID );
double getMinAngle( CPoint* cp , int pNon , int qID);
double getMaxAnge( CPoint* cp , int pNon , int qID);
bool affectPartitions(CRectangle* rect, int qID, int pNo, double cangle, double minAngle, double maxAngle);
bool UserFiltering(CRectangle* rect, int qID , int mID);
bool isRkNN(CPoint* cp, int qID);
bool liesInPartition(CRectangle* rect, int qID, int pNo, double cangle);
void SLICE_BRNN_Verification( int qID);
bool cmp( const minObject* a, const minObject* b);

void SLICE_MONO( int qID, int pNum);
void SLICE_MONO_Verification( int qID);

//=====================SIX Region ========================//
void SIX_BRNN(int qID);
void SIX_MONO( int qID );
void computeSixRegion( int qID);
void SIX_MONO_Refinement( int qID);
bool SixRegionPrune(CRectangle* rect, int qID );
bool checkInPartitions(CRectangle* rect, int qID, int pNo,double cangle) ;
void updateSixRegion(CPoint* cp, int qID , int mID);
void SIX_BRNN_Refinement( int qID);



//======================FINCH===============//

void FINCH_MONO(int qID);
void FINCH_computeSR(int qID);
void FINCH_MONO_refinement(int qID);
bool pointIsInSR(CPoint* p, int qID);

void FINCH_computeConvexHull(int qID);
void FINCH_BRNN(int qID);
void FINCH_BRNN_refinement(int qID);
bool prunedBySR(CRectangle* R, int qID);


//======================TPL=================//
void TPL_MONO(int qID);
void TPL_MONO_OPT(int qID);
void TPL_BRNN_pruningPhase(int qID);
void TPL_MONO_pruningPhase(int qID);

void TPL_MONO_refinement( int qID);
void TPL_Refinement_Round( int qID);
int getCounterOfRec(CRectangle* rec, int qID);
void TPL_BRNN(int qID);
void TPL_BRNN_OPT(int qID);

void TPL_BRNN_refinement(int qID);
void TPL_BRNN_refinement_OLD( int qID);
void MD_addInfObj(int objID, CPoint* point, int qID);  //???
CNode* findNext( int qID);
void MD_addCands(int objID, CPoint* point, int qID );
void TPL_BRNN_Verification( int qID);
void TPL_Pruning_Original( int qID);

double K_Trim_TPL( CRectangle* rec , int qID);

bool cmpHilbertValue( const infobj* s1, const infObj* s2);
int cmpHilbertValue2(const void * element1 , const void * element2);
bool cmpObjID(const infObj* s1, const infobj* s2) ;

void emptyVPolygon(int qID);

void FindRoot( int mid );



int getLowerUserIO( int qID);
int getLowerFacilityIO( int qID);
void ComputeIOBound( int qID);

bool FINCH_prunedByCircles(CRectangle* rec, int qID);
