#include "rtree.h"
#include "mystack.h"
#include <assert.h>
// #include "utility.h"

#include "leafnode.h"
#include <float.h>

#ifndef CAPACITY
#define CAPACITY 2
#endif

using namespace std;


CRTree::CRTree(char* treeName, long pageSize, double fillFactor,
               double nearMinimumOverlapFactor, double splitFactor,
               double reinsertFactor)
    : m_pageSize(pageSize),
    m_fillFactor(fillFactor),
    m_nearMinimumOverlapFactor(nearMinimumOverlapFactor),
    m_splitFactor(splitFactor),
    m_reinsertFactor(reinsertFactor)
{
  CRectangle::initMinMaxCoord();
  if ( pageSize == 0 ){
    m_nodeCapacity = CAPACITY;

    long nodeLen = 0;
    long entryLen = 0;
    CIndexNode::getBytesLength(nodeLen, entryLen);
    long pageSize_Index = entryLen * (m_nodeCapacity + 1) + nodeLen;

    CLeafNode::getBytesLength(nodeLen, entryLen);
    long pageSize_Leaf = entryLen * (m_nodeCapacity + 1)+ nodeLen;

    m_pageSize = max(pageSize_Index, pageSize_Leaf);

    m_file = new CDiskFile(treeName, m_pageSize);
  }
  else{
    calNodeCapacity();
    // open file in saveTree(), otherwise too many fstream may crash system
    m_file = new CDiskFile(treeName, pageSize);
  }
  m_memTree=true;
}

CRTree::CRTree(char* treeName)
{
  CRectangle::initMinMaxCoord();
  m_file = new CDiskFile(treeName);
  m_file->open(false);
  readHeader();

  m_memTree=false;

  loadRoot();
  m_file->close();
}

CRTree::~CRTree(void)
{
  delete m_root;
  delete m_file;
}

long CRTree::getNodeCapacity(void)
{
  return m_nodeCapacity;
}

double CRTree::getFillFactor(void)
{
  return m_fillFactor;
}

double CRTree::getNearMinimumOverlapFactor(void)
{
  return m_nearMinimumOverlapFactor;
}

double CRTree::getSplitFactor(void)
{
  return m_splitFactor;
}

double CRTree::getReinsertFactor(void)
{
  return m_reinsertFactor;
}

CNode* CRTree::getRoot(void)
{
  return m_root;
}

void CRTree::initRoot(void)
{
  m_root = new CLeafNode(this);
  // m_root->m_mbr = new CRectangle();
}

void CRTree::insertData(long id, double* coord, double prob)
{
  CRectangle* mbr = new CRectangle(coord);
  insertData(id, mbr, prob);
}

void CRTree::insertData(long id, double* low, double* high, double prob)
{
  CRectangle* mbr = new CRectangle(low, high);
  insertData(id, mbr, prob);
}

void CRTree::insertData(long id, CRectangle* mbr, double prob)
{
  bool* overflow = new bool[m_root->m_level];
  for (long i = 0; i < m_root->m_level; ++ i)
  {
    overflow[i] = false;
  }

  CStack<CNode*> path;

  CNode* node = m_root->chooseSubtree(mbr, 0, path);

  CEntry* e = new CEntry(-id, mbr, 1, prob);

  node->insertData(e, path, overflow);

  delete[] overflow;
}

void CRTree::insertData(CEntry* e, long level, bool* overflow)
{
  CStack<CNode*> path;
  CNode* node = m_root->chooseSubtree(e->m_mbr, level, path);
  node->insertData(e, path, overflow);
}

void CRTree::saveTree(void)
{
  m_file->open(true);
  allocateNodeId();

  vector<CNode*> vct;
  vct.push_back(m_root);

  CNode* curNode;

  long len = 0;

  this->writeHeader();

  while (!vct.empty())
  {
    curNode = vct[0];
    vct.erase(vct.begin());

    char* buf;
    curNode->toBytes(len, &buf);
    m_file->writeEntry(curNode->m_id, len, buf);

    delete []buf;


    if (!(curNode->isLeaf()))
    {
      for (long i = 0; i < curNode->m_size; ++ i)
      {
        vct.push_back((CNode*)(curNode->getEntry(i)));
      }
    }
  }
  m_file->close();
}

CNode* CRTree::loadRoot( )
{
  long id = 1;

  char* buf;
  m_file->readEntry(id, &buf);

  long level = -1;
  memcpy(&level, buf, sizeof(long));

  CNode* node;
  if (level == 0) // is leaf
  {
    node = new CLeafNode(this);
  }
  else  // is index
  {
    node = new CIndexNode(level, this);
  }

  node->fromBytes(buf);
  node->m_rtree = this;

  delete[] buf;

  m_root = node;

  return m_root;
}

void CRTree::loadChildren(CNode* currNode, CEntry** entries)
{
    CStack< CNode* > stk;
    long count = 0;

    stk.push(currNode);
    CNode* node;

    CIndexNode* index = NULL;
    CLeafNode* leaf = NULL;
    CNode* child = NULL;

    while (!stk.empty())
    {
        node = stk.top();
        stk.pop();
    
        if (node->isLeaf())
        {
            leaf = (CLeafNode*)node;
            for (int i = 0; i < leaf->getNumChild() ; i++)
            {
                entries[count] = leaf->getEntry( i );
                count ++;
            }
        }
        else
        {
            index = (CIndexNode*)node ;
            for (int i = 0; i < index->getNumChild(); i++)
            {
                child = (CNode*) index->getEntry(i);
                child->load(); 
                stk.push(child);
            }
        }
    }
}


void CRTree::dump(ostream& os, CNode* node)
{
  CStack<CNode*> stkNode;
  CStack<long> stkLevel;

  stkNode.push(node);
  stkLevel.push(0);

  CNode* curNode;
  long level = 0;

  while (!stkNode.empty())
  {
    curNode = stkNode.top();
    stkNode.pop();
//     if (!curNode->isLeaf())
//     {
      curNode->load();
//     }
    level = stkLevel.top();
    stkLevel.pop();

    for (long i = 0; i < level - 1; ++ i)
    {
      os << "    ";
    }
    if (level > 0)
    {
      os << "|== ";
    }
//     curNode->dump(os);
    
    //added by Bin  
    os << -curNode->m_id << " " << curNode->m_level << " ";
    curNode->m_mbr->dump(os);
    os << " " << curNode->m_dataCnt << " " << curNode->m_p;

    if (curNode->isLeaf())
    {
      for (long i = 0; i < curNode->m_size; ++ i)
      {
        for (long j = 0; j < level; ++ j)
        {
          os << "    ";
        }
        os << "|-- ";

        os << -curNode->getEntry(i)->m_id << " ";
        curNode->getEntry(i)->m_mbr->dump(os);
        
        //added by Bin
        os << " " << curNode->getEntry(i)->m_dataCnt 
            << " " << curNode->getEntry(i)->m_p;
      }
    }
    else
    {
      level ++;

      for (long i = 0; i < curNode->m_size; ++ i)
      {
        stkNode.push((CNode*)(curNode->getEntry(i)));
        stkLevel.push(level);
      }
    }
  }
}

void CRTree::loadTree()
{
  CEntry** entries;    
  int m_nOjbs = this->getRoot()->getAggCnt();
    
    // load global R tree
  entries = new CEntry*[ m_nOjbs ];
  this->loadChildren( this->getRoot() , entries );
//   dump(cout,this->getRoot());
  delete []entries;
}
void CRTree::allocateNodeId(void)
{
  /*
    here, I assume that each page is big enough for a node, and
    normally, a page stores one node only
  */
  vector<CNode*> vct;
  vct.push_back(m_root);

  long id = 1;  // root's id is 1
  CNode* curNode;

  while (!vct.empty())
  {
    curNode = vct[0];
    vct.erase(vct.begin());

    curNode->m_id = id;

    id ++;

    if (!curNode->isLeaf())
    {
      for (long i = 0; i < curNode->m_size; ++ i)
      {
        vct.push_back((CNode*)(curNode->getEntry(i)));
      }
    }
  }
}

void CRTree::calNodeCapacity(void)
{
  long nodeLen = 0;
  long entryLen = 0;
  // CLeafNode::getBytesLength(nodeLen, entryLen);
  CIndexNode::getBytesLength(nodeLen, entryLen);
  long cIndex = long(double(m_pageSize - nodeLen) / (double)entryLen);

  CLeafNode::getBytesLength(nodeLen, entryLen);
  long cLeaf = long(double(m_pageSize - nodeLen) / (double)entryLen);

  // m_nodeCapacity= 5;

  m_nodeCapacity = min( cLeaf, cIndex );
  
//   m_nodeCapacity = 2;
}

long CRTree::getHeaderSize(void)
{
  long len = sizeof(long) +     // page size
             sizeof(long) +       // node capacity
             sizeof(double) +  // fillFactor
             sizeof(double) +  // nearMinimumOverlapFactor
             sizeof(double) +  // splitFactor
             sizeof(double);     // reinsertFactor

  return len;
}

void CRTree::readHeader(void)
{
  char* buf;

  m_file->readEntry(0, &buf);

  char* ptr = buf;

  memcpy(&m_pageSize, ptr, sizeof(long));
  ptr += sizeof(long);

  memcpy(&m_nodeCapacity, ptr, sizeof(long));
  ptr += sizeof(long);

  memcpy(&m_fillFactor, ptr, sizeof(double));
  ptr += sizeof(double);

  memcpy(&m_nearMinimumOverlapFactor, ptr, sizeof(double));
  ptr += sizeof(double);

  memcpy(&m_splitFactor, ptr, sizeof(double));
  ptr += sizeof(double);

  memcpy(&m_reinsertFactor, ptr, sizeof(double));
  ptr += sizeof(double);

  delete[] buf;
}

void CRTree::writeHeader(void)
{
  long len = getHeaderSize();

  char* buf = new char[len];
  char* ptr = buf;

  memcpy(ptr, &m_pageSize, sizeof(long));
  ptr += sizeof(long);

  memcpy(ptr, &m_nodeCapacity, sizeof(long));
  ptr += sizeof(long);

  memcpy(ptr, &m_fillFactor, sizeof(double));
  ptr += sizeof(double);

  memcpy(ptr, &m_nearMinimumOverlapFactor, sizeof(double));
  ptr += sizeof(double);

  memcpy(ptr, &m_splitFactor, sizeof(double));
  ptr += sizeof(double);

  memcpy(ptr, &m_reinsertFactor, sizeof(double));
  ptr += sizeof(double);

  long id = 0;
  m_file->writeEntry(id, len, buf);

  delete []buf;
}


long CRTree::getChildren(const double* coord , vector<CEntry*> &children )
{
  CStack< CNode* > stk;

  CNode* node = m_root;
  stk.push( node );
  long num=0;

  CIndexNode * index = NULL;
  CLeafNode * leaf = NULL;
  CNode * child = NULL;

  children.clear();

  while (!stk.empty())
  {
    node = stk.top();
    stk.pop();

    switch (CRectangle::relativePos(node->getRect(), coord))
    {
    case 1:
    case 2:
    case 4:
      // nothing to do
      break;
    case 3:
    case 5:
    case 6:
      if ( node->isLeaf() )
      {
        leaf = ( CLeafNode* ) node;
        // check the child one by one.
        for ( int i=0; i < leaf->getNumChild() ; i++ )
        {
          CEntry *pEnt = leaf->getEntry( i );
          if ( CRectangle::relativePos( pEnt->getRect()->getLow() , coord  ) ==
3 )
          {
            children.push_back( pEnt );
          }
        }
      }
      else
      {
        index = ( CIndexNode* ) node ;
        for ( int i=0; i< index->getNumChild(); i ++ )
        {
          CNode *pEnt = (CNode*) index->getEntry( i );
          // child = loadNode( pEnt->m_id );
          pEnt->load();
// #ifdef R_STATIC_IO
//           if ( ! m_memTree )
//             if ( pEnt->isLeaf() )
//               CUtility::leafIO++;
//             else
//               CUtility::indexIO++;
// #endif
          stk.push( pEnt );
        }
      }
      break;
    }

  }

  return num;

}


long CRTree::getChildrenCnt( const double* coord )
{
  CStack< CNode* > stk;

  CNode* node = m_root;
  stk.push( node );
  long num=0;

  CIndexNode * index = NULL;
  CLeafNode * leaf = NULL;
  CNode * child = NULL;

  while (!stk.empty())
  {
    node = stk.top();
    stk.pop();

    switch (CRectangle::relativePos(node->getRect(), coord))
    {
    case 1:
    case 2:
    case 4:
      // nothing to do
      break;
    case 3:
    case 5:
      if ( node->isLeaf() )
      {
        leaf = ( CLeafNode* ) node;
        // check the child one by one.
        for ( int i=0; i < leaf->getNumChild() ; i++ )
        {
          CEntry *pEnt = leaf->getEntry( i );
          if ( CRectangle::relativePos( pEnt->getRect()->getLow() , coord  ) ==
3 )
          {
            num++;
          }
        }
      }
      else
      {
        index = ( CIndexNode* ) node ;
        for ( int i=0; i< index->getNumChild(); i ++ )
        {
          CNode *pEnt = (CNode*) index->getEntry( i );
          // child = loadNode( pEnt->m_id );
          pEnt->load();
          stk.push( pEnt );
        }
      }
      break;
    case 6:
      num += node->getAggCnt();
      break;
    }

  }

  return num;


}

long CRTree::getChildrenCnt( const double* coord , long MAX )
{
  return 0;
}

long CRTree::getParentCnt( const double* coord , long max )
{
  CStack< CNode* > stk;
  CNode* node = m_root;

  long num=0;

  CIndexNode * index = NULL;
  CLeafNode * leaf = NULL;
  CNode * child = NULL;

  stk.push( node );

  while (!stk.empty())
  {
    node = stk.top();
    stk.pop();

    switch (CRectangle::relativePos(node->getRect(), coord))
    {
    case 4:
    case 5:
    case 6:
      // nothing to do
      break;
    case 2:
    case 3:
      if ( node->isLeaf() )
      {
        leaf = ( CLeafNode* ) node;
        // check the child one by one.
        for ( int i=0; i < leaf->getNumChild() ; i++ )
        {
          CEntry *pEnt = leaf->getEntry( i );
          if ( CRectangle::relativePos( pEnt->getRect()->getLow() , coord  ) ==
1 )
          {
            num++;
            if ( num >= max ) break;
          }
        }
      }
      else
      {
        index = ( CIndexNode* ) node ;
        for ( int i=0; i< index->getNumChild(); i ++ )
        {
          CNode *pEnt = (CNode*) index->getEntry( i );
          // child = loadNode( pEnt->m_id );
          pEnt->load();
// #ifdef R_STATIC_IO
//           if ( ! m_memTree )
//             if ( pEnt->isLeaf() )
//             {
//               CUtility::leafIO++;
//             }
//             else
//               CUtility::indexIO++;
// #endif
          stk.push( pEnt );
        }
      }
      break;
    case 1:
      num += node->getAggCnt();
      if ( num >= max ) break;
      break;
    }

  }

  return num;

}

bool checkMBR( CRectangle * rect1 , CRectangle * rect2 )
{
  for ( int i=0; i< DIMENSION; i++ )
  {
    if (  rect1->getLow()[i] < rect2->getLow()[i])
    {
      cerr << "LOW : " << rect1->getLow()[i] << " : " << rect2->getLow()[i] <<
endl;
      return false;
    }
    if ( rect1->getHigh()[i] > rect2->getHigh()[i])
    {
      cerr << "HIGH : " << rect1->getHigh()[i] << " : " << rect2->getHigh()[i]
<< endl;
      return false;
    }
  }

  return true;
}

bool Verify( CNode* pEnt )
{

  if ( pEnt->isLeaf() )
  {
    CEntry * e;
    for ( int i=0; i< pEnt->getNumChild(); i++ )
    {
      e = pEnt->getEntry( i );
      if ( !checkMBR( e->getRect() ,  pEnt->getRect() ) )
      {
        cerr << "Error MBR (Leaf )" << endl;
        return false;
      }
    }

  }
  else
  {
    CNode * e;
    for ( int i=0; i< pEnt->getNumChild(); i++ )
    {
      e = (CNode*) pEnt->getEntry( i );
      if ( ! e->isLeaf() )
        e->load();
        
        if ( !checkMBR( e->getRect() ,  pEnt->getRect() ) )
        {
        cerr << "Error MBR (Index)" << endl;
        return false;
        }

      bool ret = Verify( (CNode*) e );

      if ( !ret ) return ret;

    }
  }
  
  return true;
}

bool CRTree::verifyMBR()
{
  return Verify( m_root );
}


void CRTree::traverseTree()
{
  CStack< CNode* > stk;
  CNode* node = m_root;
  CIndexNode * index = NULL;
  CLeafNode * leaf = NULL;
  CNode * child = NULL;
  stk.push( node );
  CPoint* p;
  
  while (!stk.empty())
  {
    node = stk.top();
    stk.pop();
    if ( node->isLeaf() )
    {
      leaf = ( CLeafNode* ) node;
      for ( int i=0; i < leaf->getNumChild() ; i++ )
      {
        
        CEntry *pEnt = leaf->getEntry( i );
        p = new CPoint(pEnt->getRect()->m_low);
        long id = pEnt->m_id;
        cout << id << "     ";
        p->print();
        cout << endl;
      }
    }
    else
    {
      index = ( CIndexNode* ) node ;
      for ( int i=0; i< index->getNumChild(); i ++ )
      {
        CNode *pEnt = (CNode*) index->getEntry( i );
        
         cout << "pushing in stack "<< endl;
        pEnt->load(); 
        stk.push( pEnt );  
        
      }

    }
  }
}

int CRTree::getFanout()
{
  CStack< CNode* > stk;
  CNode* node = m_root;
  CIndexNode * index = NULL;
  CLeafNode * leaf = NULL;
  CNode * child = NULL;
  stk.push( node );
  CPoint* p;
  
  
  int totChild = 0;
  int totNodes = 0;
  while (!stk.empty())
  {
    node = stk.top();
    stk.pop();
    if ( node->isLeaf() )
    {
      leaf = ( CLeafNode* ) node;
      
      totNodes++;
      totChild += leaf->getNumChild( );
    
    }
    else
    {
      index = ( CIndexNode* ) node ;
      
      totNodes++;
      totChild += index->getNumChild( );
      
//       cout << index->getNumChild() << endl;
      for ( int i=0; i< index->getNumChild(); i ++ )
      {
        CNode *pEnt = (CNode*) index->getEntry( i );
        
//         cout << "pushing in stack "<< endl;
        pEnt->load(); 
        stk.push( pEnt );  
        
      }

    }
  }
  
  //cout << "totChild: " << totChild << " totNodes: " << totNodes << endl;
  return totChild/totNodes;
}


void CRTree::getAllObjects(CNode* currNode, list<CNode*> *allObjects)
{
  CStack< CNode* > stk;
  long count = 0;

  stk.push(currNode);
  CNode* node;

  CIndexNode* index = NULL;
  CLeafNode* leaf = NULL;
  CNode* child = NULL;

  while (!stk.empty())
  {
    node = stk.top();
    stk.pop();
    
    if (node->isLeaf())
    {
      leaf = (CLeafNode*)node;
      for (int i = 0; i < leaf->getNumChild() ; i++)
      {
        child = (CNode*)leaf->getEntry( i );
        allObjects->push_back(child);
        
        count ++;
      }
    }
    else
    {
      index = (CIndexNode*)node ;
      for (int i = 0; i < index->getNumChild(); i++)
      {
        child = (CNode*) index->getEntry(i);
        child->load(); 
        stk.push(child);
      }
    }
  }
}

