#ifndef RTREE_H
#define RTREE_H

#include <iostream>

#include "node.h"
#include "rectangle.h"
#include "diskfile.h"

using namespace std;

// #define R_STATIC_IO

class CEntry;
class CNode;
class CIndexNode;
class CLeafNode;

class CRTree
{
private:
  // m_root: the root node
  CNode* m_root;

  // m_nodeCapacity: the number of entries in a leaf node
  // this value is automatically calculated according to m_pageSize
  

  double m_fillFactor;
  double m_nearMinimumOverlapFactor;
  double m_splitFactor;
  double m_reinsertFactor;

public:

  CDiskFile* m_file;
  long m_nodeCapacity;

  long m_pageSize;

  bool m_memTree;

public:
  // create a new tree
  CRTree(char* treeName, long pageSize, double fillFactor = 0.5,
         double nearMinimumOverlapFactor = 32, double splitFactor = 0.4,
         double reinsertFactor = 0.3);

  // load a tree
  CRTree(char* treeName);

  virtual ~CRTree(void);

  long getNodeCapacity(void);

  double getFillFactor(void);
  double getNearMinimumOverlapFactor(void);
  double getSplitFactor(void);
  double getReinsertFactor(void);

  CNode* getRoot(void);

  // initRoot: initialize root, e.g., create/load the root node
  void initRoot(void);

  // insertData: insert a new data point
  void insertData(long id, double* coord, double prob);

  // insertData: insert a new data rectangle
  void insertData(long id, double* low, double* high, double prob);

  // insertData: the main entrance for insert a new data
  void insertData(long id, CRectangle* mbr, double prob);

  // saveTree: save the current rtree to disk
  void saveTree(void);

  // loadTree: load whole tree from disk
  void loadTree(void);

  // return number of points covered by coord
  long getChildrenCnt( const double* coord  ) ;
  long getChildren(const double* coord , vector<CEntry*> &children );

  // return number of points covered by coord,
  // return MAX if cnt >= MAX
  long getChildrenCnt( const double* coord , long MAX ) ;

  long getParentCnt( const double* coord , long MAX ) ;

  // loadNode: load the root from disk
  // other entry is loaded by node->load();
  // 'id' is the block id of the node
  CNode* loadRoot();
  void dump(ostream& os, CNode* node);
  bool verifyMBR();
  
  void loadChildren(CNode* currNode, CEntry** entries);
  
  // Print the entries in the tree to see if the tree generated seems ok
  void traverseTree();
  
  // Load all the objects in the node
  void getAllObjects(CNode* currNode, list<CNode*> *allObjects);
  
  // traverse the whole R-Tree and by looking at numChild in each node, compute the fanout
  int getFanout();
  
private:
  // calNodeCapacity: calcualte node capacity
  void calNodeCapacity(void);

  // insertData: do reinsert a data point
  void insertData(CEntry* e, long level, bool* overflow);

  // allocateNodeId: allocate id of each node according to page size
  void allocateNodeId(void);

  long getHeaderSize(void);
  void readHeader(void);
  void writeHeader(void);

  friend class CNode;
  friend class CIndexNode;
  friend class CLeafNode;
};

#endif
